import click
import asyncio
from . import KunYu 
from .exceptions import AnimeNotFoundError, CharacterNotFoundError
S = "\x1b[38;5;33m"
E = "\x1b[0m"
VA = "\x1b[38;5;203m"

@click.group()
def cli():
    """AnimeScraper - CLI to fetch and search anime/character details from MyAnimeList."""
    pass

@click.command()
@click.argument('anime_name')
def search_anime(anime_name: str):
    """Search for an anime by name."""
    try:
        async def run():
            async with KunYu() as scraper:
                print(f"\x1b[38;5;45m> Searching for anime\x1b[0m: \x1b[38;5;44m{anime_name}\x1b[0m\n")
                anime = await scraper.search_anime(anime_name)

                click.echo(f"{S}Title{E}: {VA}{anime.title}{E}")
                click.echo(f"{S}Japanese Title{E}: {VA}{anime.japanese_title}{E}")
                click.echo(f"{S}Type{E}: {VA}{anime.anime_type}{E}")
                click.echo(f"{S}Episodes{E}: {VA}{anime.episodes}{E}")
                click.echo(f"{S}Duration{E}: {VA}{anime.duration}")
                click.echo(f"{S}Status{E}: {VA}{anime.status}{E}")
                click.echo(f"{S}Rating{E}: {VA}{anime.rating}{E}")
                click.echo(f"{S}Genres{E}: {VA}{' '.join(anime.genres)}{E}")
                click.echo(f"{S}Themes{E}: {VA}{' '.join(anime.themes)}{E}")
                click.echo(f"{S}Premiered{E}: {VA}{anime.premiered}{E}")
                click.echo(f"{S}Aired{E}: {VA}{anime.aired}{E}")
                click.echo(f"{S}Score{E}: {VA}{anime.stats.score}{E}")
                click.echo(f"{S}Scored by{E}: {VA}{anime.stats.scored_by} people{E}")
                click.echo(f"{S}Popularity{E}: {VA}{anime.stats.popularity}{E}")
                click.echo(f"{S}Ranked{E}: {VA}{anime.stats.ranked}{E}")
                click.echo(f"{S}Synopsis{E}: \x1b[38;5;35m{anime.synopsis[:300]}...\x1b[0m")

        asyncio.run(run())
    except AnimeNotFoundError as e:
        click.echo(f"Error: {e}", err=True)



@click.command()
@click.argument('anime_id')
def get_anime(anime_id: str):
    """Get anime details by anime ID."""
    try:
        async def run():
            async with KunYu() as scraper:
                print(f"\x1b[38;5;45m> Fetching anime details for ID: \x1b[0m: \x1b[38;5;44m{anime_id}\x1b[0m")

                anime = await scraper.get_anime(anime_id)

                genres = " ".join(anime.genres)
                click.echo(f"{S}Title{E}: {VA}{anime.title}{E}")
                click.echo(f"{S}Japanese Title{E}: {VA}{anime.japanese_title}{E}")
                click.echo(f"{S}Type{E}: {VA}{anime.anime_type}{E}")
                click.echo(f"{S}Episodes{E}: {VA}{anime.episodes}{E}")
                click.echo(f"{S}Duration{E}: {VA}{anime.duration}")
                click.echo(f"{S}Status{E}: {VA}{anime.status}{E}")
                click.echo(f"{S}Rating{E}: {VA}{anime.rating}{E}")
                click.echo(f"{S}Genres{E}: {VA}{genres}{E}")
                click.echo(f"{S}Themes{E}: {VA}{' '.join(anime.themes)}{E}")
                click.echo(f"{S}Premiered{E}: {VA}{anime.premiered}{E}")
                click.echo(f"{S}Aired{E}: {VA}{anime.aired}{E}")
                click.echo(f"{S}Score{E}: {VA}{anime.stats.score}{E}")
                click.echo(f"{S}Scored by{E}: {VA}{anime.stats.scored_by} people{E}")
                click.echo(f"{S}Popularity{E}: {VA}{anime.stats.popularity}{E}")
                click.echo(f"{S}Ranked{E}: {VA}{anime.stats.ranked}{E}")
                click.echo(f"{S}Synopsis{E}: \x1b[38;5;35m{anime.synopsis[:300]}...\x1b[0m")
        asyncio.run(run())

    except AnimeNotFoundError as e:
        click.echo(f"Error: {e}", err=True)



@click.command()
@click.argument('character_id')
def get_character(character_id: str):
    """Get character details by character ID."""
    try:
        async def run():
            async with KunYu() as scraper:

                print(f"\x1b[38;5;45m> Fetching character details for ID: \x1b[0m: \x1b[38;5;44m{character_id}\x1b[0m\n")
                character = await scraper.get_character(character_id)

                click.echo(f"{S}Name{E}: {VA}{character.name}{E}")
                click.echo(f"{S}Japanese Name{E}: {VA}{character.japanese_name}")
                for key, value in character.about.items():
                    click.echo(f"{S}{key}{E}: {VA}{value}{E}")
                click.echo(f"{S}Description{E}: \x1b[38;5;35m{character.description[:300]}...{E}")  # Limit the description

        asyncio.run(run())
    except CharacterNotFoundError as e:
        click.echo(f"Eror: {e}")



@click.command()
@click.argument('character_name')
def search_character(character_name: str):
    """Get character details by character name."""
    try: 
        async def run():
            async with KunYu() as scraper:

                print(f"\x1b[38;5;45m> Searching for character\x1b[0m: \x1b[38;5;44m{character_name}\x1b[0m\n")
                character = await scraper.search_character(character_name)

                click.echo(f"{S}Name{E}: {VA}{character.name}{E}")
                click.echo(f"{S}Japanese Name{E}: {VA}{character.japanese_name}")
                for key, value in character.about.items():
                    click.echo(f"{S}{key}{E}: {VA}{value}{E}")
                click.echo(f"{S}Description{E}: \x1b[38;5;35m{character.description[:300]}...{E}")  # Limit the description
        asyncio.run(run())

    except CharacterNotFoundError as e:
        click.echo(f"Eror: {e}")

# Add all commands to the CLI
cli.add_command(search_anime)
cli.add_command(search_character)
cli.add_command(get_anime)
cli.add_command(get_character)


if __name__ == '__main__':
    cli()




import re
from bs4 import BeautifulSoup
from typing import Dict, List
from ._model import Anime, AnimeCharacter, AnimeStats, Character
from rapidfuzz import process, fuzz
from .exceptions import CharacterNotFoundError

def _parse_anime_data(html: str)-> Anime:

    soup = BeautifulSoup(html, "html.parser")
    anime_id = soup.find("input", attrs={"name": "aid"}).attrs["value"] # type: ignore

    title = soup.find("h1", "title-name h1_bold_none").text # type: ignore
    jap_title = get_span_text(soup, "Japanese") 
    eng_title = get_span_text(soup, "English")
    anime_type = get_span_text(soup, "Type")
    episodes = get_span_text(soup, "Episodes")
    duration = get_span_text(soup, "Duration")
    status = get_span_text(soup, "Status")
    aired = get_span_text(soup, "Aired")
    premiered = get_span_text(soup, "Premiered")
    studios = get_span_text(soup, "Studios")
    rating = get_span_text(soup, "Rating")
    synopsis = soup.find_all("p", attrs={'itemprop': 'description'})[0].text

    theme = soup.find("span", "dark_text", string="Theme:")  # type: ignore
    if theme == None:
        theme = soup.find("span", "dark_text", string="Themes:")
    theme_list = [i.string for i in theme.parent.find_all("a")] if theme else ["N/A"] # type: ignore 
    
    genres = soup.find("span", "dark_text", string="Genres:")
    if genres == None:
        genres = soup.find("span", "dark_text", string="Genre:")
    genres_list = [a.string for a in genres.parent.find_all("a")] if genres else ["N/A"]  # type: ignore 

    anime_stats = get_anime_stats(soup)
    anime_characters = _anime_characters(soup)

    return Anime(
        anime_id,
        title,
        eng_title,
        jap_title,
        anime_type,
        episodes,
        status,
        aired,
        duration,
        premiered,
        rating,
        synopsis,
        genres_list,
        studios,
        theme_list,
        anime_stats,
        anime_characters
    )



def get_anime_stats(soup: BeautifulSoup)-> AnimeStats:
    score = soup.find("span", attrs={"itemprop": "ratingValue"}).text #type: ignore
    scored_by = soup.find("span", attrs={"itemprop": "ratingCount"}).text #type: ignore
    popularity = get_span_text(soup, "Popularity")
    members = get_span_text(soup, "Members")
    favorites = get_span_text(soup, "Favorites")
    ranked = soup.find("span", "numbers ranked").strong.text # type: ignore
    return AnimeStats(
        score, 
        scored_by, 
        ranked, 
        popularity, 
        members, 
        favorites
    )


def _anime_characters(soup)-> List[AnimeCharacter]:

    tables = soup.find("div", "detail-characters-list clearfix").find_all("table", attrs={"width":"100%"}) #type: ignore    
    characters = []

    for table in tables:
        character = table.find("h3", "h3_characters_voice_actors")
        characters.append([
            get_id(character.a.get("href")), # character id
            remove_coma(character.a.text),   # character name
            character.parent.small.text,     # character role
            get_voice_actor(table)           # voice actor information
            ])
    
    return [AnimeCharacter(*c) for c in characters]


def get_voice_actor(table)-> Dict[str, str]:
    td_tag = table.find("td", "va-t ar pl4 pr4")
    if td_tag:
        return {

            "id": get_id(td_tag.a.get("href")), # voice actor MAL ID 
            "name": remove_coma(td_tag.a.text), # voice actor name
            "role": td_tag.small.text,
            "url": td_tag.a.get("href") #voice actor information page url
        }
    return {"id": "N/A","name": "N/A", "role": "N/A", "url": "N/A"}



def remove_coma(name: str):
    return "".join(name.split(",")) # Name has coma. ex: "Togashi, Yuuta"


def get_id(url):
    """Get ID of character/people/anime fron URL"""
    return url.split("https://myanimelist.net/")[1].split("/")[1]


def get_span_text(soup: BeautifulSoup, info_name: str)->str:
    if soup.find("span", "dark_text", string=f"{info_name}:"):
        return soup.find("span", "dark_text", string=f"{info_name}:").parent.text.strip().split(f"{info_name}:")[-1].strip() # type: ignore 
    else:
        return "N/A"


def parse_anime_search(html)-> List:
    soup = BeautifulSoup(html, "html.parser")
    data = [
        [tag.text, tag.get("href")] for tag in soup.find_all("a", "hoverinfo_trigger fw-b fl-l")
    ]

    return data

def parse_character_search(html)-> tuple:
    soup = BeautifulSoup(html, "html.parser")
    tag = soup.find("a")
    name = tag.text #type: ignore
    url = tag.get("href") #type: ignore
    return (name, url)


def parse_the_character(html):
    # Htto response code: 200 though invalid id/name was given 
    # Check html before parsing
    if '<div class="badresult">Invalid ID provided.</div>' in html:
        raise CharacterNotFoundError("The MAL Character id is Invalid")

    soup = BeautifulSoup(html, "html.parser")
    url = soup.find("meta", {"property":"og:url"}).get("content") #type: ignore
    img = soup.find("meta", {"property": "og:image"}).get("content") #type: ignore
    match = re.search(r"Member Favorites:\s*([\d,]+)", soup.get_text())
    # Remove commas for integer conversion
    member_favorites = match.group(1).replace(",", "")  #type: ignore
    start = '<h2 class="normal_header" style="height: 15px;">'
    end = '<div class="normal_header">'

    html = html.split(start)[1].split(end)[0]
    soup = BeautifulSoup(html, "html.parser")

    for spoiler in soup.find_all("div", "spoiler"):
        spoiler.decompose()

    cleaned_text = str(soup)
    pattern = r".*?:\s*<br\s*/?>\n?"  # Matches "Text: <br>" or "Text:<br/>" followed by an optional newline
    cleaned_text = re.sub(pattern, "", cleaned_text)
    soup = BeautifulSoup(cleaned_text, "html.parser")


    # removing <br> to prepare `about` dictionary
    clean_content = "".join(cleaned_text.split("<br/>"))
    soup_for_texts = BeautifulSoup(clean_content, "html.parser")

    # spliting to get the names and character `about`
    lines = soup_for_texts.get_text("\n").splitlines()
    name, japanese_name = lines[0], lines[1]

    # dictionary may contain age, height, Weight etc.
    about = {
        key: value.strip() for line in lines if ":" in line and len(line) <40 and "(Source" not in line for key, value in [line.split(':')]
    }
    # spliting the soup for character description
    pure_texts = soup_for_texts.get_text(strip=True)

    #getting last item of dict to split description
    k, v = next(reversed(about.items()))
    # last_item = f'{k}: {v}'
    description = "".join(pure_texts.split(f'{k}: {v}')[1:]).strip()

    return Character(
        get_id(url),
        name,
        japanese_name,
        about,
        description,
        img, #type: ignore
        member_favorites,
        url #type: ignore
    )

def typ(url: str)-> str:
    """Returns TYPE from url anime, character etc."""
    if "cat=" in url:
        return "".join(url.split("=")[-1])
    else:
        return "".join(url.split("https://myanimelist.net/")[1].split("/")[0])

def normalize(text)-> str:
    return re.sub(r"[^a-zA-Z0-9\s]", "", text).lower()


def get_close_match(query, lists):
    return process.extractOne(normalize(query), lists, scorer=fuzz.ratio)





import aiosqlite
import json
import sqlite3
async def _initialize_database(db_path):
        """
        Initializes the SQLite database with necessary tables.
        """
        async with aiosqlite.connect(db_path) as db:
            await db.execute("""
                CREATE TABLE IF NOT EXISTS anime (
                    id TEXT PRIMARY KEY,
                    data TEXT
                )
            """)
            await db.execute("""
                CREATE TABLE IF NOT EXISTS character (
                    id TEXT PRIMARY KEY,
                    data TEXT
                )
            """)
            await db.commit()


async def _get_from_cache(db, table: str, key: str):
        async with db.execute(f"SELECT data FROM {table} WHERE id = ?", (key,)) as cursor:
            row = await cursor.fetchone()
            if row:
                return json.loads(row[0])  # Return deserialized JSON
        return None



async def _store_in_cache(db, table: str, key: str, value: dict):
        async with db.execute(f"SELECT 1 FROM {table} WHERE id = ?", (key,)) as cursor:
            row = await cursor.fetchone()
            if row:
                return  # Skip if the data already exists
        await db.execute(f"INSERT INTO {table} (id, data) VALUES (?, ?)", (key, json.dumps(value)))
        await db.commit()



def _start_database(db_path):
        """
        Initializes the SQLite database with necessary tables.
        """
        with sqlite3.connect(db_path) as db:
            cursor =  db.cursor()
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS anime (
                    id TEXT PRIMARY KEY,
                    data TEXT
                )
            """)
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS character (
                    id TEXT PRIMARY KEY,
                    data TEXT
                )
            """)
            db.commit()


def _from_cache(db, table: str, key: str):
        cursor = db.execute(f"SELECT data FROM {table} WHERE id = ?", (key,))
        row = cursor.fetchone()
        if row:
            print(row[0])
            return json.loads(row[0])  # Return deserialized JSON
        return None


def _store_cache(db, table: str, key: str, value: dict):
        cursor = db.execute(f"SELECT 1 FROM {table} WHERE id = ?", (key,))
        row = cursor.fetchone()
        if row:
            return  # Skip if the data already exists
        db.execute(f"INSERT INTO {table} (id, data) VALUES (?, ?)", (key, json.dumps(value)))
        db.commit()




"""
Scraper: A module to fetch and parse data from MyAnimeList.

This module contains the **KunYu()** class, which acts as the primary entry point
for interacting with `Myanimelist <https://myanimelist.net/>`__ data, as well as helper classes and functions.
"""

from typing import List, Optional
import aiohttp
from ._model import Anime, Character
from .async_malscraper import MalScraper

class KunYu:
    """
    The main class for interacting with MyAnimeList data.

    This class acts as the entry point for users of the library.

    """



    def __init__(
            self, 
            use_cache: bool = False, 
            db_path: str = "cache.db",
            max_requests: int = 5,
            per_second: int = 1,
            timeout: int = 10
    ) -> None:
        """
        Initial method.

        Args:
            use_cache (bool): If data should be cached. (Default: False)
            db_path: (str): The path of the database. (Default: cache.db)
            max_requests (int): The number requests to make at `per_second` seconds. (Default: 5)
            per_second (int): number of seconds `max_requests` can be made. (Default: 1)

        """


        self.shared_session: Optional[aiohttp.ClientSession] = None
        self.Scraper = MalScraper(
            session=self.shared_session,
            use_cache=use_cache,
            db_path=db_path,
            max_requests=max_requests,
            per_second=per_second,
            timeout=timeout
        )
    




    async def __aenter__(self):
        # These Headers are needed in oder to get porper response from MAL.
        self.shared_session = aiohttp.ClientSession(
            headers = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
    "Accept-Language": "en-US,en;q=0.5",
    "Connection": "keep-alive",
    "Upgrade-Insecure-Requests": "1"
            })

        return self


    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.shared_session:
            await self.shared_session.close()




    async def search_anime(self, anime_name: str)-> Anime:
        """
        Fetches and Returns Anime details from myanimelist.

        Args:
            anime_name (str): Name of the anime you want to search.
        
        Returns:
            Anime: Returns Anime object with anime details.
        """

        async with self.Scraper as scraper:
            anime = await scraper.search_anime(anime_name)
            return anime




    async def search_character(self, character_name: str)-> Character:
        """
        Fetches and Returns Character details from myanimelist.

        Args:
            character_name (str): Name of the Character you want to search.
        
        Returns:
            Character: Returns Character object with the character details.
        """

        async with self.Scraper as scraper:
            character = await scraper.search_character(character_name)
            return character



    async def get_anime(self, anime_id: str)->Anime:
        """
        Fetches anime details from MyAnimeList.

        Args:
            anime_id (str): The ID of the anime.

        Returns:
            Anime: An object containing anime details.
        """

        async with self.Scraper as scraper:
            anime = await scraper.get_anime(anime_id)
            return anime




    async def get_character(self, character_id: str)-> Character:
        """
        Fetches character details from MyAnimeList.

        Args:
            character_id (str): The ID of the character.

        Returns:
            Character: An object containing character details.
        """

        async with self.Scraper as scraper:
            character = await scraper.get_character(character_id)
            return character


    async def search_batch_anime(self, anime_names: List)-> List[Anime]:
        """
        Fetches multiple anime in batch.

        Args:
            anime_names (List): List of anime names.

        Returns:
            List[Anime]: Returns a list of Anime class object with anime details.

        """

        async with self.Scraper as scraper:
            batch_anime = await scraper.search_batch_anime(anime_names)
            return batch_anime 



    async def search_batch_character(self, character_names: List)-> List[Character]:
        """
        Fetches multiple characters in batch.

        Args:
            character_names (List): List of characters name.

        Returns:
            List[Character]: Returns a list of Character class object with characters details.

        """

        async with self.Scraper as scraper:
            batch_characters = await scraper.search_batch_character(character_names)
            return batch_characters
        



import httpx
from typing import Optional, List 
from urllib.parse import quote 
from concurrent.futures import ThreadPoolExecutor
import sqlite3


from .exceptions import (
    AnimeNotFoundError, 
    CharacterNotFoundError, 
    NetworkError
)
from ._cache_utils import _start_database, _from_cache, _store_cache
from ._parse_anime_data import (
    _parse_anime_data,
    parse_anime_search, 
    parse_the_character,
    parse_character_search,
    get_id,
    get_close_match,
    normalize
)

from ._model import Anime, Character


class SyncMalScraper():

    BASE_URL = "https://myanimelist.net"
    ANIME = 10 
    CHARACTER = 20

    def __init__(
        self, 
        client: Optional[httpx.Client],
        use_cache: bool,
        db_path: str,
        timeout: int
        ) -> None:
        self.client = client
        self.own_client = client is None
        self.use_cache = use_cache
        self.db_path = db_path
        self.db: sqlite3.Connection | None = None
        self.timeout = timeout

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1"
    }

    def __enter__(self):

        if not self.client:
            self.client = httpx.Client(headers=self.headers)

        if self.use_cache:
            _start_database(self.db_path)
            self.db = sqlite3.connect(self.db_path)
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.client and self.own_client:
            self.client.close()
        if self.db:
            self.db.close()


    def _fetch(self, url: str, req: int, query: str)-> str:
        
        if not self.client:
            raise ValueError("session is not initialised. Use `with SyncMalScraper` for proper initialisation")
        try:
            response = self.client.get(url, timeout=self.timeout)
            if response.status_code == 404 and self.ANIME == req:
                raise AnimeNotFoundError(query)
            elif response.status_code == 404 and self.CHARACTER == req:
                raise CharacterNotFoundError(query)
            return response.text

        except httpx.NetworkError as e:
            raise NetworkError(f"A NetworkError error occurred {e}")


    def get_anime(self, anime_id: str)->Anime:
        """
        Fetch and parse anime details.

        Args:
            anime_id (str): The MyAnimeList ID of the anime.

        Returns:
            Anime: An object containing detailed anime information.
        """
        if self.use_cache:
            cached_data = _from_cache(self.db, "anime", anime_id)
            if cached_data:
                return Anime._from_dict(cached_data)

        url = f"{self.BASE_URL}/anime/{anime_id}"

        html = self._fetch(url, self.ANIME, anime_id)
        anime =  _parse_anime_data(html)
        if self.use_cache:
            _store_cache(self.db, "anime", anime_id, Anime._to_dict(anime))
        return anime



    def get_character(self, character_id: str)-> Character:
        """
        Fetch and parse character details.

        Args:
            character_id (str): The MyAnimeList ID of the character.

        Returns:
            Character: An object containing detailed character information.
        """
        
        if self.use_cache:
            cached_data = _from_cache(self.db, "character", character_id)
            if cached_data:
                return Character._from_dict(cached_data)

        url = f"{self.BASE_URL}/character/{character_id}"
        html = self._fetch(url, self.CHARACTER, character_id)
        character = parse_the_character(html)
        if self.use_cache:
            _store_cache(self.db, "character", character_id, Character._to_dict(character))
           
        return character



    def search_anime(self, query: str)-> Anime:

        url = f"{self.BASE_URL}/anime.php?q={quote(query)}&cat=anime"
        html = self._fetch(url, self.ANIME, query)
        start = '<table border="0" cellpadding="0" cellspacing="0" width="100%">'
        end = '<td class="borderClass bgColor1" valign="top" width="50">'

        # spliting and getting table contents remove useless codes
        anime_lists = "".join(html.split(start)[1].split(end, 9)[1:-1])
        # ((anime name, anime url)) tuple
        allanime = parse_anime_search(anime_lists)
        animeNames = tuple((x[0] for x in allanime))

        matched = get_close_match(query, animeNames)
        # if match rate > 50 return matched anime else first anime from list
        index = matched[2] if matched[1] > 50 else 0
        url = allanime[index][1]
        return self.get_anime(get_id(url))



    def search_character(self, query: str)-> Character:
        """
        Search Character by name in myanimelist.net

        Args:
            query (str): The name of the Character.

        Returns:
            Character: A Character object with The Character Details.
        """

        url = f"{self.BASE_URL}/character.php?q={query}&cat=character"
        html = self._fetch(url, self.CHARACTER, query)
        start = '<table border="0" cellpadding="0" cellspacing="0" width="100%">'
        end = '</table>'

        # getting all the Character row from table of search results
        table = html.split(start)[1].split(end)[0].split('width="175">', 8)
        chars = tuple((parse_character_search(x) for x in table))

        names = tuple((normalize(x[0]) for x in chars))

        # if match rate higher than 50 return match else first char
        matched = get_close_match(query, names)
        index = matched[2] if matched[1] > 50 else 0
        
        url = chars[index][1]

        return self.get_character(get_id(url))



    def search_batch_anime(self, anime_names: List[str])-> List[Anime]:
        with ThreadPoolExecutor(max_workers=5) as threat:
            results = threat.map(self.search_anime, anime_names)
            return [result for result in results]


    def search_batch_character(self, characters_name: List[str])-> List[Character]:

        characters_lists = [self.search_character(name) for name in characters_name]
        return characters_lists



"""
AnimeScraper Exception classes
"""
class AnimeScraperError(Exception):
    """Base class for all custom exceptions in AnimeScraper."""
    pass


class AnimeNotFoundError(AnimeScraperError):
    """Raised when an anime is not found on MyAnimeList."""
    def __init__(self, anime_name: str):
        super().__init__(f"\x1b[38;5;124mAnime with id/name\x1b[0m \x1b[38;5;220m'{anime_name}'\x1b[0m \x1b[38;5;124mnot found on MyAnimeList.\x1b[0m")


class CharacterNotFoundError(AnimeScraperError):
    """Raised when a character is not found on MyAnimeList."""
    def __init__(self, character_name: str):
        super().__init__(f"\x1b[38;5;124mCharacter with id/name\x1b[0m '{character_name}'\x1b[0m \x1b[38;5;124mnot found on MyAnimeList.\x1b[0m")


class NetworkError(AnimeScraperError):
    """Raised when there is a network-related issue."""
    def __init__(self, message: str = "\x1b[38;5;124mA network error occurred.\x1b[0m"):
        super().__init__(message)

